-- PEPA module ---------------------------------------------------------
module PepaParser (
	     Component (..), parsePepa, parsePepaWithLog
	     ) where


--   import Char (isAlpha, isAlphaNum, isSpace)
   import PepaTypes (Identifier)
   import Utils ((?),numberLines)
   import LogResult (LogResult, newLR)

-- Lexer definitions begin ---------------------------------------------

   type Activity = Identifier
   type Rate = Identifier

   data Component =  
	   PREFIX ((Activity, Rate), Component)
	 | CHOICE (Component, Component)
	 | COOP   (Component, Component, [Activity])
	 | HIDING (Component, [Activity])
	 | VAR    Identifier
	 | CONSTS (Identifier, Component, Component)

   instance Show Component where
      show (CONSTS (ci, c1, c2))   = concat ["# ", ci, " = ", show c1, ";\n", show c2]
      show (CHOICE (c1, c2))       = concat [show c1, " + ", show c2]
      show (PREFIX ((act, rt), c)) = concat ["(", act, ", ", rt, ").",show c]
      show (HIDING (c, ids))       = concat ["(", show c, ") \\ {", foldr1 (\x -> \y -> x++", "++y) ids, "}"]
      show (VAR ids)               = ids
      show (COOP (c1, c2, ids))    = concat ["(",show c1, " <", (null ids) ? ("", foldr1 (\x -> \y -> x++", "++y) ids),"> ", show c2, ")"]


   data Token = 
       Ident Identifier
     | Symbol Char
       deriving Show

   isIdChar :: Char -> Bool
   isIdChar c = isAlphaNum c || c `elem` "'_*-" 

   isPepaSym :: Char -> Bool
   isPepaSym c = c `elem` "/{}<>().,=+#;"

   isPepaComment :: Char -> Bool
   isPepaComment = ('%'==)

   ident :: [Char] -> Token
   ident l = Ident (reverse l)

   getword :: [Char] -> [Char] -> [Token]
   getword l [] = [ident l]
   getword l (a:x) =
	 if isIdChar a
	 then getword (a:l) x
	 else ident l : analyse (a:x)    

   analyse :: [Char] -> [Token]
   analyse [] = []
   analyse (a:x) =
	 if isSpace a then analyse x else 
	 if isPepaSym a then Symbol a : analyse x else 
	 if isAlpha a then getword [a] x else
	 if isPepaComment a then analyse $ unlines $ tail $ lines (a:x) else
	    error ("PepaParser: Unrecognised token " ++ (a:x))

-- Lexer definitions end -----------------------------------------------

-- Parser definitions begin --------------------------------------------

-- Parser tools.  Adapted from Chris Reade's splendid book 
--                ``Elements of Functional Programming''

-- Utilities 
   pair :: a -> b -> (a, b)
   pair a b = (a, b)

   consonto :: [a] -> a -> [a]
   consonto x a = a:x
  
   cons :: (a, [a]) -> [a]
   cons (x, xs) = x : xs
-- End utilities

   type Parser a = [Token] -> Maybe (a, [Token])

   infixr 4 <&>
   infixr 3 <|>
   infix  0 `modify`

   (<|>) :: Parser a -> Parser a -> Parser a
   parser1 <|> parser2 = \ s -> case (parser1 s) of
                                  Nothing -> parser2 s
                                  x       -> x

   modify :: Parser a -> (a -> b) -> Parser b
   parser `modify` f = \ s -> case (parser s) of
                              Nothing     -> Nothing
                              Just (x, y) -> Just (f x, y)

   (<&>) :: Parser a -> Parser b -> Parser (a, b)
   parser1 <&> parser2 = \ s -> case (parser1 s) of
                                  Nothing       -> Nothing
				  Just (x1, s1) -> (parser2 `modify` (pair x1)) s1

   emptyseq :: Parser [a]
   emptyseq s = Just ([], s)

   optional :: Parser a -> Parser [a]
   optional pr = (pr `modify` (consonto []))
                  <|> emptyseq

   parserseq :: Parser a -> Parser [a]
   parserseq pr = seqpr where 
                          seqpr = (pr <&> seqpr `modify` cons) <|> emptyseq

   seqwith :: (Parser a, Parser b, Parser c) -> Parser d -> Parser [d] 
   seqwith (front, sep, back) pr = 
        front <&> optional items <&> back `modify` (concat . fst . snd) where
          sep_pr = sep <&> pr `modify` snd
          items  = pr  <&> parserseq sep_pr `modify` cons

   parserList :: [Parser a] -> Parser [a]
   parserList []        = emptyseq
   parserList (pr:rest) = pr <&> (parserList rest) `modify` cons

   alternatives :: [Parser a] -> Parser a
   alternatives []        = \ x -> Nothing
   alternatives (pr:rest) = pr <|> alternatives rest
------------------------------------------------------------------------    
-- Basic parsers begin -------------------------------------------------

   variable :: Parser String
   variable ((Ident x) : s)   = Just (x, s)
   variable _                 = Nothing 

   literal :: Char -> Parser Char
   literal a ((Symbol x) : s) = if a == x then Just (x, s) else Nothing
   literal _ _                = Nothing
    
-- Basic parsers end ---------------------------------------------------
------------------------------------------------------------------------    

------------------------------------------------------------------------    
-- A parser for PEPA ---------------------------------------------------
   unparenth :: (a,(b,c)) -> b
   unparenth (bra, (e, ket)) = e

   bracedseq :: Parser a -> Parser [a]
   bracedseq = seqwith (literal '{', literal ',', literal '}')

   angledseq :: Parser a -> Parser [a]
   angledseq = seqwith (literal '<', literal ',', literal '>')

   actlist :: Parser [Identifier]
   actlist = bracedseq variable

   cooplist :: Parser [Identifier]
   cooplist = angledseq variable

   def  s   = ((adef <&> optional (literal ';' <&> def)   `modify` opt_consts)
                <|> agenta                                                   ) s
 
   adef s   = (optional (literal '#')  <&> variable  <&>
                literal '='            <&> agenta           `modify` mk_const) s

   agenta s = (agentb <&> optional (literal '/' <&> actlist) 
                                                           `modify` mk_hiding) s

   agentb s = (agentc <&> optional (literal '+' <&> agentb) 
                                                             `modify` mk_plus) s

   agentc s = (agent <&> optional (cooplist <&> agentc) 
                                                             `modify` mk_coop) s

   agent s  = ((literal '('    <&> variable    <&>
                literal ','    <&> variable    <&>
                literal ')'    <&> literal '.' <&> 
                agent                                      `modify` mk_prefix)
               <|> (literal '(' <&> agenta      <&> literal ')' 
                                                           `modify` unparenth)
               <|> (variable                               `modify` mk_agent)) s

   mk_agent s                               = VAR s

   mk_const (hash, (i, (eq, e)))            = (i, e)

   mk_hiding (p, [])                        = p
   mk_hiding (p, [(slash, l)])              = HIDING (p, l)
   mk_hiding other                          = error "PepaParser: hiding"

   mk_plus (p, [])                          = p
   mk_plus (p, [(plus, q)])                 = CHOICE (p, q)
   mk_plus other                            = error "PepaParser: plus"

   mk_coop (p, [])                          = p
   mk_coop (p, [(l, q)])                    = COOP (p, q, l)
   mk_coop other                            = error "PepaParser: cooperation"

   mk_prefix (bra, (alpha, (comma, (rate, (ket, (dot, p))))))
                                            = PREFIX ((alpha, rate), p)

   opt_consts ((i, e1), [(oper, e2)])       = CONSTS (i, e1, e2)
   opt_consts other                         = error "PepaParser: constants"
   
-- End parser for PEPA -------------------------------------------------
------------------------------------------------------------------------    


-- Parser tools end ----------------------------------------------------
------------------------------------------------------------------------  
  
   lit :: Token -> String
   lit (Symbol ';') = ";\n\t"
   lit (Symbol '=') = " = "
   lit (Symbol '+') = " + "
   lit (Symbol ',') = ", "
   lit (Symbol '<') = " <"
   lit (Symbol '>') = "> "
   lit (Symbol c)   = [c]
   lit (Ident s)    = s

   report :: Maybe (a,[Token]) -> a
   report Nothing        = error "PepaParser: ill-formed PEPA model definition"
   report (Just (c, [])) = c
   report (Just (c, x))  = error (concat ("PepaParser: Unparsed :-\n" : (map lit x)))

   parsePepa :: String -> Component
   parsePepa = report . def . analyse

   parsePepaWithLog :: String -> LogResult Component
   parsePepaWithLog = (newLR "PepaParser" analyseComponent) . parsePepa

   analyseComponent :: Component -> String
   analyseComponent cpt = concat [show $ countConsts cpt, " PEPA constant(s) found\n",
			    "Model successfully parsed:\n", numberLines $ lines $ show cpt] where
			  countConsts (CONSTS (_,_,c)) = 1 + (countConsts c)
			  countConsts _ = 0

-- PEPA module ends ----------------------------------------------------
------------------------------------------------------------------------    
